SystemOrganization addCategory: #'Prolog-Compiler'!


Compiler subclass: #LCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LCompiler class
	instanceVariableNames: ''!

Decompiler subclass: #LDecompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LDecompiler class
	instanceVariableNames: ''!

Parser subclass: #LParser
	instanceVariableNames: 'prologMethodSelector prologMethodArgument cuts temporaries statements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Compiler'!

LParser class
	instanceVariableNames: ''!


!LCompiler commentStamp: 'spfa 6/20/2008 12:39' prior: 0!
Compiler for Prolog class tree!
]style[(30)!

!LDecompiler commentStamp: 'spfa 2/26/2017 13:04' prior: 0!
Decompiler for Prolog class tree

Does not produce actual working code for Prolog methods because of the selector dispatching system; see usages of #selectorForPredicate:!
]style[(170)!

!LParser commentStamp: 'bolot 12/13/2002 15:14' prior: 0!
RE: safeGuardReturnBlock
Makes sure that the method returns, even if no match occurs.

RE: queryUndefined
Disabled for Prolog methods (because undefined Prolog variables are unbounds).

RE: prologMethodSelector
Can be used to determine if this is a smalltalk method or prolog method.

RE: Smalltalk guys
Strings, symbols, numbers, and literal points (45@65) are recognized.

RE: evaluation
Support for nicer, more Prolog-like syntax for evaluation is done through direct use of LCompiler.
Example: LCompiler evaluate: 'Family new :? father(''John'', x).'
(see Family comment for current situation)

This class and other related classes in this category were originally written by or are based on the work by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.
We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!
]style[(1032)!

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/3/2022 23:59'!
addTemp: aTerm
	"Decide whether aTerm should be declared as a temporay."
	| var term |

	self assert: aTerm first isUppercase.
	term := aTerm withFirstCharacterDownshifted.
	
	var := encoder encodeVariable: term
		ifUnknown: [encoder bindTemp: term].

	(temporaries includes: var) ifTrue: [^var].

	temporaries add: var.
	statements addFirst: 
		(AssignmentNode new
			variable: var
			value: (self send: #newTerm: to: 'LogicRef' with: {aTerm})).
	^var! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
argumentList: aList
	^ aList isConsCell
		ifTrue: [self arguments: aList]
		ifFalse: [aList]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
arguments: aList
	"Answer a MessageNode for predicate arguments."
	| receiver bList |

	aList ifNil: [^ nil] "zero argument".

	receiver := self send: #car:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head}.

	aList cdr ifNil: [^ receiver] "single argument". 

	aList cdr isConsCell ifFalse: ["[x | y]"
		^ self send: #car:cdr:
			to: (aList isRelation ifTrue: ['Relation'] ifFalse: ['PrologList'])
			with: {self argumentList: aList head . self argumentList: aList tail}].

	"more than one arguments"
	bList := aList.
	[(bList := bList tail) isConsCell] whileTrue:
		[receiver := self send: #addLast:
						to: receiver
						with: {self argumentList: bList head}].
	bList ifNotNil:
		[^ self send: #append:
					to: receiver
					with: {self argumentList: bList}].
	^ receiver! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/11/2022 23:35'!
atomicTerm: aNumber
	
	^ self atomicTerm: aNumber afterOperator: nil delimiters: self termDelimiters! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/13/2022 00:19'!
atomicTerm: aNumber afterOperator: operator delimiters: delimiters

"The code here is highly heuristic and very difficult to make sense of.
TO BE REFACTORED
but with care, because it may break parsing that happens only during compilation of very specifc Prolog methods, such as the ones using DCGs. Test extensively after any change!!"


	| term token hend |

	hend := hereEnd.

	(self match: #exclamationMark) ifTrue: [^ self predCut]. 

	(delimiters includes: hereType)
		ifTrue: [self expected: 'Term'].

	token := LogicUnbound. "to be used as the void value - see all tests below"

	"POSSIBLE EXPRESSION - token is made of terms within {...}"
	(hereType == #leftBrace)
		ifTrue: [token := self bracedTerms].

	"LIST - token is a list of terms"
	(hereType == #leftBracket)
		ifTrue: [token := self listTerms].

	"PARENTHESIS - token has the form (...terms...)"
	(token == LogicUnbound) ifTrue: [
		(self match: #leftParenthesis) ifTrue: [
			token := self atomicTerm: 1.
			(self match: #rightParenthesis) ifFalse: [
				self advance.
				^ (self commaParsingOf: 
					(self atomicTerm: 1 afterOperator: nil 
							delimiters: (self termDelimiters copyWithout: #comma)) 
						topForms) asPrologList]]].

	term := here. "the following attempts will advance the parser, so note the term now"
	term = $, ifTrue: [term := here asString]. 


	"ATOMIC - token is a Point"
	(token == LogicUnbound) ifTrue: [
		(hereType == #number and: [token == #@])
			ifTrue:
				[self advance; advance.
				term := term @ here.
				self advance.
				token := term]].

	"ATOMIC - token is a number, string or literal"
	(token == LogicUnbound) ifTrue: [
		(hereType == #string or: [hereType == #number or: [hereType == #literal]])
			ifTrue: 
				[self advance.
				token := term]].


	"VARIABLE - token is an underscore"
	(token == LogicUnbound) ifTrue: [
		(hereType == #word or: [hereType == #binary] or: [hereType == #comma]  or: [hereType == #colonDash])
			ifFalse: [
				(self match: #leftArrow) "should be #underscore"
					ifTrue: [token := self send: #new to: 'LogicDontCare']
					ifFalse: [self notify: 'undefined token: ', token]]].

	"If we have the token but the term is not ended, then the syntax needs the other tokens to be parsed before we can make any sense of it - so redirect the parsing accordingly"
	(token == LogicUnbound) ifFalse: [
		(delimiters includes: hereType) ifTrue: [^ token].
		aNumber = 100 ifTrue: [^ token].
		^ self predicateFromOperatorsAfter: token delimiters: (delimiters copyWithout: #comma)].

	"we still do not have the token at this point"
	self advance.

	"we have a symbol before an opening parens
	if there is a gap between them, it may be a prefix operator"
	((hereType = #leftParenthesis)
		and: [hereMark - hend > 1])
			ifTrue: [
				token := (PrologOperator named: term asSymbol) ifNil: [self error].
				token isPrefix ifTrue: [
				^ self predicateFromOperatorsAfter: token]].

	"we have a symbol before an opening parens
	if the token is not in first position and not in position of being an operator argument
then it cannot be a functor and so it is an operator"
	((hereType = #leftParenthesis)
		and: [aNumber = 100]
		and: [operator isNil or: [operator isPostfix]])
			ifTrue: [
				(PrologOperator named: term asSymbol) ifNotNil: [:op | ^ op]]. "well it may be a functor after all"

	"we have a symbol before an opening parens
	 if it is not an operator, it must be a functor"
	(self match: #leftParenthesis)
		ifTrue: [
			token := self predicate: term asSymbol.
			(delimiters includes: hereType) ifTrue: [^ token].
			^ self predicateFromOperatorsAfter: token 
					delimiters: (delimiters copyWithout: #comma) ].
	

	"VARIABLE - token is a named variable"
	(term first isLetter and: [term first isUppercase]) ifTrue: [
		token := self addTemp: term.
		(delimiters includes: hereType) ifTrue: [^ token].
		aNumber = 100 ifTrue: [^ token].
		^ self predicateFromOperatorsAfter: token delimiters: delimiters].

"at this point the token will be an ATOM
the issue is whether we make it a symbol or an operator"

	(delimiters includes: hereType) ifTrue: [
		^ PrologOperator ifAnyFor: term]. "well, could also be a postfix operator"

	aNumber < 100 ifTrue: [
		token := PrologOperator ifAnyFor: term.
					
		aNumber = 1 ifTrue: [
			^ self predicateFromOperatorsAfter: token 
				delimiters: (delimiters copyWithout: #comma)].
	
		^ self predicateFromOperatorsAfter: token delimiters: delimiters].

	^ PrologOperator ifAnyFor: term! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 16:03'!
bracedTerms
	

	self notYetImplemented! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 15:30'!
buildPredicateFrom: anArray

	^ self buildPredicateFrom: anArray ensurePrecedenceBelow: nil! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/11/2022 22:46'!
buildPredicateFrom: anArray ensurePrecedenceBelow: aNumber

	| topop ti prec |

	anArray size = 1 ifTrue: [^ anArray anyOne].

	topop := (anArray select: [:ea | ea class == PrologOperator]) detectMax: #precedence.

	prec := topop precedence.

	(aNumber notNil and: [prec = aNumber])
		ifTrue: [self error: 'Ambiguous operator associativity - parens required']. 

	ti := anArray indexOf: topop.

	(topop type = #xfx) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)
					ensurePrecedenceBelow: prec. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)
					ensurePrecedenceBelow: prec} 
				asPrologRelation].

	(topop type = #yfx) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)
					ensurePrecedenceBelow: prec. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)} 
				asPrologRelation].

	(topop type = #xfy) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1). 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)
					ensurePrecedenceBelow: prec} 
				asPrologRelation].

	(#(fx fy) includes: topop type) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: ti+1 to: anArray size)} 
				asPrologRelation].

	(#(xf yf) includes: topop type) ifTrue: [
		^ {topop operator. 
			self buildPredicateFrom: (anArray copyFrom: 1 to: ti-1)} 
				asPrologRelation].

	self error: 'Badly defined operator: ', topop operator! !

!LParser methodsFor: 'as yet unclassified' stamp: 'spfa 10/13/2022 09:44'!
commaParsingOf: anObject

	(anObject isPrologList and: [anObject car == #COMMA]) 
		ifTrue: [^ (self commaParsingOf: anObject second), (self commaParsingOf: anObject third)].

	anObject == #COMMA ifTrue: [^ #()].

	anObject isArray ifFalse: [^ {anObject}].

	^ Array streamContents: [:str |
		anObject do: [:ea | 
			str nextPutAll: (self commaParsingOf: ea)]]! !

!LParser methodsFor: 'as yet unclassified' stamp: 'spfa 10/12/2022 00:31'!
compileDCG: aPrologList

	"Definite clause grammar 
	--> syntactic sugar"

	| block relations tmps tmp term args terms |

	(self match: #period) ifFalse: [self expected: 'Period']. 

	args := aPrologList head isSymbol 
				ifTrue: [#()] 
				ifFalse: [aPrologList head cdr topForms].

	terms := aPrologList tail.

	relations := Array streamContents: [:str |

		terms topForms do: [:t |
			term := t.
			term isSymbol ifTrue: [term := Relation car: term].
			term ifNil: [str nextPut: (#terminal -> nil)]
				ifNotNil: [
					(term isPrologList and: [term isRelation not]) 
						ifTrue: [
							term topForms "terminals" do: [:form |
								str nextPut: (#terminal -> form)]] 
						ifFalse: [str nextPut: term]].]].

	tmps := (1 to: relations size) collect: [:n |
		self addTemp: 'S', (n-1)].

	tmp := self addTemp: 'S'.

	relations := relations collectWithIndex: [:ea :n |
		| s0 s1 |
		n = relations size 
			ifTrue: [s0 := tmps last. s1 := tmp]
			ifFalse: [s0 := tmps at: n. s1 := tmps at: n+1].
		ea class == Association "terminals"
			ifTrue: [
				(ea value isNil and: [args isEmpty]) 
					ifTrue: [{#unify . s0.  s1} asPrologRelation]
					ifFalse: [{#cccc . s0. ea value . s1} asPrologRelation]] 
			ifFalse: [(ea topForms, {s0. s1}) asPrologRelation]].

	block := self predicateSeriesFromRelations: relations.

	^ self send: #unify:with:then:
		to: 'self'
		with: {
				self send: #key to: prologMethodArgument.
				self arguments: (args, {tmps first. tmp}) asPrologList.
				(hereType == #doIt and: [cuts not])
						ifTrue: [self returnBlock: block] 
						ifFalse: block
				}! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/13/2022 00:00'!
compileRule: predicate block: aBlockNode 

	predicate head == prologMethodSelector
		ifFalse: [self notify: 'predicate name different'].

	(self match: #period) ifFalse: [self expected: 'Period']. 

	^ self send: #unify:with:then:
		to: 'self'
		with: {
				self send: #key to: prologMethodArgument.
				self arguments: predicate tail.
				(hereType == #doIt and: [cuts not])
						ifTrue: [self returnBlock: aBlockNode]  "optimize the stack" 
						ifFalse: aBlockNode
				}! !

!LCompiler class methodsFor: 'evaluating' stamp: 'spfa 10/13/2022 16:11'!
deResultify: anArray

	^ anArray select: [:ea | ea class ~~ LogicUnification] thenCollect: #logicResult! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 00:42'!
decompile: selector in: aClass method: aMethod using: aConstructor

"wrong - many non-predicate selectors starting the prologKeyword are indeed unary
 #logicValue, etc"
[
	((selector beginsWith: Logic prologKeyword)
		and: [selector isUnary])
	"restore the one-argument form of selector"
	ifTrue: [^ super decompile: (selector, ':') asSymbol in: aClass method: aMethod using: aConstructor].
].

	^ super decompile: selector in: aClass method: aMethod using: aConstructor

! !

!LCompiler class methodsFor: 'accessing' stamp: 'spfa 11/18/2009 18:10'!
decompilerClass

	^LDecompiler! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/6/2022 18:34'!
expandTerm: term
	"
Transcript visible showln: term -> term class.

	term = 42 ifTrue: [^ 54].

	(term class == LogicRef and: [term term = 'X']) ifTrue: [term term: 'Z'].
"
	^ term! !

!LCompiler class methodsFor: 'evaluating' stamp: 'spfa 10/13/2022 15:43'!
generatorOnQueryTo: aLogic for: eString withBindings: aBoolean

	^ Generator on: [:g |
		[self evaluate: 'self yourself :? ', eString for: aLogic logged: false]
			on: PrologAnswerStep 
			do: [:ex | 
				g yield: (aBoolean ifTrue: [ex answer] 
					ifFalse: [self deResultify: ex answer]).
				ex resume]]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/13/2022 00:24'!
goalSeries
	"Do the work for method statements by
	 adding each goal to statements."
	" (Prolog new :? goal1(x,y), goal2(z,y)) do: [:i | ...] "

	^self arguments: self topLevelTerms
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 9/30/2022 01:07'!
initialize

	super initialize.
	typeTable := TypeTable copy "safer"! !

!LParser methodsFor: 'error correction' stamp: 'spfa 11/18/2009 18:10'!
interactive

	^ (requestor isKindOf: REPLMorph) not
			and: [super interactive]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/6/2022 14:32'!
listTerms
	"Answer a list of predicate terms."
	| term aList |
	self advance.
	(self match: #rightBracket) ifTrue: [^ nil].
	term := self expandTerm: (self term).
	aList := PrologList car: term.
	[
		(self match: #verticalBar) ifTrue: [
				aList append: self term.
				(self match: #rightBracket) ifTrue: [^ aList].
				self expected: 'Right bracket'].
		(self match: #comma)
			ifTrue: [aList addLast: (self expandTerm: (self term))]
			ifFalse: [
				(self match: #rightBracket) ifTrue: [^aList].
				self expected: 'Right bracket']
	] repeat! !

!LParser methodsFor: 'expression types' stamp: 'spfa 10/13/2022 00:24'!
messagePart: level repeat: repeat
	"bolot 8/5/1999 18:14"
	"This is the same as Parser>>messagePart:repeat:"
	"with only ONE addition: support for #:? for calling Prolog stuff"

	| start receiver selector args precedence words keywordStart |
	[receiver := parseNode.
	(hereType == #keyword and: [level >= 3])
		ifTrue: 
			[start := self startOfNextToken.
			selector := WriteStream on: (String new: 32).
			args := OrderedCollection new.
			words := OrderedCollection new.
			[hereType == #keyword]
				whileTrue: 
					[keywordStart := self startOfNextToken + requestorOffset.
					selector nextPutAll: self advance.
					words addLast: (keywordStart to: self endOfLastToken + requestorOffset).
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 2 repeat: true.
					args addLast: parseNode].
			(Symbol hasInterned: selector contents ifTrue: [ :sym | selector := sym])
				ifFalse: [ selector := self correctSelector: selector contents
										wordIntervals: words
										exprInterval: (start to: self endOfLastToken)
										ifAbort: [ ^ self fail ] ].
			precedence := 3]
		ifFalse: [((hereType == #binary or: [hereType == #verticalBar])
				and: [level >= 2])
				ifTrue: 
					[start := self startOfNextToken.
					selector := self advance asSymbol.
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 1 repeat: true.
					args := Array with: parseNode.
					precedence := 2]
				ifFalse: [hereType == #word
						ifTrue: 
							[start := self startOfNextToken.
							selector := self advance.
							args := #().
							words := OrderedCollection with:
									(start + requestorOffset to: self endOfLastToken + requestorOffset).
							(Symbol hasInterned: selector ifTrue: [ :sym | selector := sym])
								ifFalse: [ selector := self correctSelector: selector
													wordIntervals: words
													exprInterval: (start to: self endOfLastToken)
													ifAbort: [ ^ self fail ] ].
							precedence := 1]
						ifFalse: [(hereType == #colonQuestion)
								ifTrue: [selector := #inferSmalltalk:.
										"change typeTable for Prologging"
										typeTable at: $, asciiValue put: #comma;
											at: $!! asciiValue put: #exclamationMark.
										self advance. 
										args := Array with: self goalSeries .
										"restore typeTable for Smalltalking"
										typeTable at: $, asciiValue put: (TypeTable at: $, asciiValue);
											at: $!! asciiValue put: (TypeTable at: $!! asciiValue).
										precedence := 3 "TODO: pick right precedence"]
								ifFalse: [^args notNil]]]].
	parseNode := MessageNode new
				receiver: receiver
				selector: selector
				arguments: args
				precedence: precedence
				from: encoder
				sourceRange: (start to: self endOfLastToken).
	repeat]
		whileTrue: [].
	^true! !

!LParser methodsFor: 'expression types' stamp: 'spfa 10/11/2022 23:15'!
method: doit context: ctxt

    | sap blk prim messageComment methodNode |

    "need temporaries/statements in doits - to enable unbounds"
    temporaries := OrderedCollection new.
    statements := OrderedCollection new.
    "may need to change (temporarily) for doits"
    typeTable := typeTable copy.

    "check if it's a Prolog method definition"
	((doit not and: [here beginsWith: Logic prologKeyword]) 
		and: [(here last == $:) not])
     ifTrue:
            [^self prologMethod: doit context: ctxt].

    sap := self pattern: doit inContext: ctxt.
    "sap={selector, arguments, precedence}"
  "  (sap at: 2) do: [:argNode | argNode beMethodArg]."   "... or   isArg: true"
    temporaries := self temporaries asOrderedCollection.

    messageComment := currentComment.
    currentComment := nil.
    prim := doit ifTrue: [0] ifFalse: [self primitive].
    self statements: #() innerBlock: doit.

    parseNode statements: statements , parseNode statements.

    blk := parseNode.
    doit ifTrue: [blk returnLast]
        ifFalse: [blk returnSelfIfNoOther].
    (self termDelimiters includes: hereType) ifFalse: [^self expected: 'Nothing more'].
 "   self interactive ifTrue: [self removeUnusedTemps]."

    methodNode := self newMethodNode comment: messageComment.
    ^methodNode
        selector: (sap at: 1)
        arguments: (sap at: 2)
        precedence: (sap at: 3)
        temporaries: temporaries
        block: blk
        encoder: encoder
        primitive: prim! !

!LParser methodsFor: 'expression types' stamp: 'spfa 4/15/2010 17:57'!
method: doit context: ctxt encoder: encoderToUse
	" pattern [ | temporaries ] block => MethodNode."

	encoder := encoderToUse.
	^self method: doit context: ctxt! !

!LCompiler class methodsFor: 'accessing' stamp: 'bolot 7/2/1999 17:25'!
parserClass
	"Return a parser class to use for parsing method headers."

	^LParser! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
predCut
	cuts := true.
	^ PrologList with: #primlogicCut: 
		with: (self returnBlock: (self send: #asArray to: 'answer'))
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/11/2022 10:32'!
predicate: aSymbol

	| result |

	(aSymbol = 'iss' or: [aSymbol = 'consult'])
		ifFalse: [^ self structure: aSymbol].

	"special cases: involve parsing smalltalk code"

	result := Relation with: aSymbol.
	result addLast: (self expandTerm: self term). "the Prolog side, first argument"
	(self match: #comma)
		ifFalse: [self expected: 'Term (comma?)'].
	self expression
		ifFalse: [self expected: 'Expression'].
    (self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].
	result addLast:
		(BlockNode new
			arguments: #()
			statements: {parseNode}
			returns: false
			from: encoder).

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 01:08'!
predicateFromOperatorsAfter: aTerm

	^ self predicateFromOperatorsAfter: aTerm delimiters: self termDelimiters! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 15:05'!
predicateFromOperatorsAfter: aTerm delimiters: delimiters

	| runop |

	runop := nil.
	aTerm class == PrologOperator ifTrue: [runop := aTerm].

	^ self buildPredicateFrom: 
		(Array streamContents: [:str |
			str nextPut: aTerm.
			[
				runop := str nextPut: 
					(self atomicTerm: 100 afterOperator: runop delimiters: delimiters).
				runop class == PrologOperator ifFalse: [runop := nil].
				delimiters includes: hereType
			] whileFalse]).! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/10/2022 17:44'!
predicateSeries
	| pred |
	pred := self topLevelTerm.
    ^BlockNode new
		arguments: #()
		statements: {self send: pred head
							to: 'self'
							with: {self send: #key:value:
										to: 'Association'
										with: {
											self arguments: pred tail .
											(self match: #comma)
												ifTrue: [self predicateSeries]
												ifFalse:[self send: #value
															to: prologMethodArgument
															with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/13/2022 09:29'!
predicateSeries: aCons
	| pred |
	pred := aCons car asPrologRelation.
    ^BlockNode new
		arguments: #()
		statements: {self send: pred head
							to: 'self'
							with: {self send: #key:value:
										to: 'Association'
										with: {
											self arguments: pred tail .
											aCons cdr 
												ifNotNil: [self predicateSeries: aCons cdr]
												ifNil: [self send: #value
															to: prologMethodArgument
															with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/7/2022 23:56'!
predicateSeriesFromRelations: anArray
	| pred |
	pred := anArray first.

    ^BlockNode new
		arguments: #()
		statements: {
			self send: (Logic makeLogicSelector: pred head)
				to: 'self'
				with: {self send: #key:value:
							to: 'Association'
							with: {
								self arguments: pred tail .
								anArray size > 1 
									ifTrue: [self predicateSeriesFromRelations: anArray allButFirst]
									ifFalse:[self send: #value
												to: prologMethodArgument
												with: #()]}}}
		returns: false
		from: encoder! !

!LParser methodsFor: 'hack' stamp: 'spfa 11/18/2009 18:10'!
primitive

	"ugly hack for 3.8 -> 3.10 compatibility"

	^ [self pragmaPrimitives] on: Error do: [super primitive]! !

!LParser methodsFor: 'expression types' stamp: 'spfa 11/18/2009 18:10'!
prologMethod: doit context: ctxt

	| blk messageComment methodNode blockComment |

	prologMethodSelector := (self pattern: doit inContext: ctxt) first.

	typeTable at: $, asciiValue put: #comma.
	typeTable at: $!! asciiValue put: #exclamationMark.

	(prologMethodArgument := encoder bindArg: #assoc).
	" prologMethodArgument  beMethodArg. "  "... or   isArg: true"
	messageComment := currentComment.
	currentComment := nil.
	" self statements: #() innerBlock: doit. "
	parseNode := BlockNode new
				arguments: #()
				statements: (self ruleSeries, self safeGuardReturnStatements)
				returns: doit
				from: encoder.
	blockComment := currentComment.
	currentComment := nil.
	parseNode comment: blockComment.
	blk := parseNode.
	hereType == #doIt ifFalse: [^self expected: 'Nothing more'].
	methodNode := MethodNode new comment: messageComment.

	^methodNode
		selector: prologMethodSelector
		arguments: {prologMethodArgument}
		precedence: 3
		temporaries: temporaries
		block: blk
		encoder: encoder
		primitive: 0! !

!LParser methodsFor: 'error correction' stamp: 'spfa 9/27/2022 21:26'!
queryUndefined
	"a lot of Prolog variables are undefined because they are unbound"

[
	"let's prevent unnecessary interrogation (unless it's a Smalltalk method)"
	prologMethodSelector ifNil: [^super queryUndefined]
]! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
returnBlock: anExpression
	"Answer a Block which returns anExpression."

	^BlockNode new
		arguments: #()
		statements: {
			ReturnNode new
				expr: anExpression
				encoder: encoder
				sourceRange: (1 to: 1)}
		returns: true
		from: encoder! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/13/2022 00:27'!
rule
	"rule:
		predicate [:- predicates] 
	or
		head --> body"

	| predicate|

	cuts := false.
	predicate := self topLevelTerm.

	"case 1) predicate :- predicates."
	(self match: #colonDash) ifTrue: [
		^ self compileRule: predicate block: (self predicateSeries: self topLevelTerms)].

	"case 2) head --> body."
	(predicate head = #logicDCG) ifTrue: [
		^ self compileDCG: (self commaParsingOf: predicate tail topForms) asPrologList].

	"case 3) predicate." 	"(alone)"
	^ self compileRule: predicate block: (self send: #value to: prologMethodArgument)
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
ruleSeries
	"Do the work for method statements by
	 adding each rule to statements."

	[statements add: self rule. source atEnd] whileFalse.
	^statements! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
safeGuardReturnStatements
	^ {
		self send: #ifTrue:ifFalse: 
			to: (self send: #isEmptyOrNil to: 'answer')
			with: {
					self returnBlock: (encoder encodeVariable: 'nil') .
					self returnBlock: (self send: #asArray to: 'answer')
					}
		}
	! !

!LDecompiler methodsFor: 'as yet unclassified' stamp: 'spfa 9/29/2022 01:24'!
send: selector super: superFlag numArgs: numArgs

	| sel |

	sel := selector.

	(selector isUnary
		and: [numArgs > 0]
		and: [selector beginsWith: Logic prologKeyword])
	ifTrue: [
	"here selector is a Prolog predicate that will be dynamically looked up in Logic>>#doesNotUnderstand. 	we 'decompile' it by giving an abstract selector, for information only."
		sel := ('logic.', 
				(selector allButFirst: Logic prologKeyword size) 					withFirstCharacterDownshifted,
				 ':') asSymbol].

	^ super send: sel super: superFlag numArgs: numArgs
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver 

	^ self send: selector to: receiver with: #()! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 11/18/2009 18:10'!
send: selector to: receiver with: arguments
	"selector is a Symbol or a Selector
	node. receiver can be
	either a String or other nodes."

	^ MessageNode new
		receiver: (receiver isString
						ifTrue: [encoder encodeVariable: receiver]
						ifFalse: [receiver])
		selector: (selector asSymbol)
		arguments: (arguments collect: [:elem |
							(elem isKindOf: ParseNode)
								ifTrue: [elem]
								ifFalse: [encoder encodeLiteral: elem]])
		precedence: 3
		from: encoder
		sourceRange: (1 to: 1)! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/11/2022 17:29'!
structure: aSymbol
	"Process the argument part of a structure
	 whose functor is aSymbol. Current token
	 is now the left paren. Answer a Relation
	 representing the structure."

	| result |

	result := Relation with: aSymbol.
	(self match: #rightParenthesis) ifTrue: [^ result].

	[result addLast: (self expandTerm: (self term)). (self match: #comma)] whileTrue.

	(self match: #rightParenthesis)
		ifFalse: [self expected: 'Right parenthesis'].

	^result! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 14:15'!
term

	^ self atomicTerm: 2! !

!LParser methodsFor: 'as yet unclassified' stamp: 'spfa 10/12/2022 13:56'!
termDelimiters

	^ #(period colonDash rightParenthesis rightBracket verticalBar comma doIt)
! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 13:52'!
topLevelTerm

	| pred term |
 
	term := (self atomicTerm: 1) asPrologRelation.

	(cue getClass binaryPredicates keys includes: term car)
		ifTrue: [pred := cue getClass selectorForBinaryPredicate: term car]
		ifFalse: [pred := term car asString].

	pred := Logic makeLogicSelector: pred.

	term car: pred.

	^ term! !

!LParser methodsFor: 'prolog parser' stamp: 'spfa 10/12/2022 15:29'!
topLevelTerms

	| pred terms cterms |
 

	terms := self atomicTerm: 1 afterOperator: nil delimiters: #(period rightParenthesis rightBracket verticalBar doIt  ).

 	cterms := (self commaParsingOf: terms).

	cterms do: [:ea | | term |

		term := ea asPrologRelation.

		(cue getClass binaryPredicates keys includes: term car)
			ifTrue: [pred := cue getClass selectorForBinaryPredicate: term car]
			ifFalse: [pred := term car asString].

		pred := Logic makeLogicSelector: pred.

		term car: pred].

	^ cterms asPrologList! !

!LParser methodsFor: 'as yet unclassified' stamp: 'spfa 10/11/2022 17:31'!
xBinary

	| startOfToken |
	tokenType := #binary.
	startOfToken := mark.

	token := String with: self step.

"HACK"
	"Allow the proper parsing of negative numbers"
	(token = '-'  
		and: [(self typeTableAt: hereChar) ~~ #xDelimiter] 
		and: [hereType ~= #number] 
		and: [(self typeTableAt: aheadChar) == #xDigit])
	ifTrue: [
		self xDigit. token := token negated. ^ self].

	[(self typeTableAt: hereChar) == #xBinary or: [(self typeTableAt: hereChar) == #verticalBar]] whileTrue:
		[(hereChar == $- and: [(self typeTableAt: aheadChar) == #xDigit])
			ifTrue: [^self ambiguousSelector: (token , '-')
					inRange: (startOfToken to: source position - 1).].
		token := token, (String with: self step)].

	token := token asSymbol! !

!LParser methodsFor: 'multi-character scans' stamp: 'spfa 11/18/2009 18:10'!
xColon		"process prolog :- and :? symbols"
	aheadChar = $- ifTrue:
		[self step.
		tokenType := #colonDash.
		self step.
		^ token := #colonDash].
	aheadChar = $? ifTrue:
		[self step.
		tokenType := #colonQuestion.
		self step.
		^ token := #colonQuestion].
	^super xColon! !

!LParser methodsFor: 'hack' stamp: 'spfa 12/21/2012 14:17'!
xLetter
	"Form a word or keyword."

	| type c |
	
	(SystemVersion current majorMinorVersion >= 'Squeak4.4')
		ifTrue: [^ super xLetter].
	
	buffer reset.
	[c := hereChar asciiValue.
	(type := typeTable at: c ifAbsent: [#xLetter]) == #xLetter or: [type == #xDigit]]
		whileTrue: ["open code step for speed"
			buffer nextPut: hereChar.
			hereChar := aheadChar.
			source atEnd
				ifTrue: [aheadChar := 30 asCharacter
					"doit"]
				ifFalse: [aheadChar := source next]].
	(type == #colon or: [type == #xColon and: [aheadChar ~= $=]])
		ifTrue: [buffer nextPut: self step.
			["Allow any number of embedded colons in literal symbols"
			(typeTable at: hereChar asciiValue ifAbsent: [#xLetter])
				== #xColon]
				whileTrue: [buffer nextPut: self step].
			tokenType := #keyword]
		ifFalse: [tokenType := #word].
	token := buffer contents.
	token isOctetString ifTrue: [token := token asOctetString].
! !
